#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <dirent.h>
#include <stddef.h>
#include <fcntl.h>
#include <string.h>

int main(void)
{
  pid_t pid;
  int status;
  char initpath[256];

#ifdef ARCH_386
  /* Work around rpm bug */
  char buffer[4096];
  char *remove_file[] = {
    "/lib/i686/libc-" VERSION ".so",
    "/lib/i686/libm-" VERSION ".so",
    "/lib/i686/libpthread-" PVERSION ".so",
    "/lib/i686/libc.so.6",
    "/lib/i686/libm.so.6",
    "/lib/i686/libpthread.so.0",
    NULL };
  struct pref {
    char *p;
    int len;
  } prefix[] = { { "libc-", 5 }, { "libm-", 5 }, { "libpthread-", 11 } };
  int i, fd;
  off_t base;
  ssize_t ret;

  for (i = 0; remove_file[i]; i++)
    unlink (remove_file[i]);
  fd = open ("/lib/i686", O_RDONLY);
  if (fd >= 0
      && (ret = getdirentries (fd, buffer, sizeof (buffer), &base))
	>= (ssize_t) offsetof (struct dirent, d_name))
    {
      for (base = 0; base + offsetof (struct dirent, d_name) < ret; )
       {
	 struct dirent *d = (struct dirent *) (buffer + base);

	 for (i = 0; i < sizeof (prefix) / sizeof (prefix[0]); i++)
	   if (! strncmp (d->d_name, prefix[i].p, prefix[i].len))
	     {
	       char *p = d->d_name + prefix[i].len;

	       while (*p == '.' || (*p >= '0' && *p <= '9')) p++;
	       if (p[0] == 's' && p[1] == 'o' && p[2] == '\0'
		   && p + 3 - d->d_name < sizeof (initpath) - strlen ("/lib/i686/"))
		 {
		   strcpy (initpath, "/lib/i686/");
		   strcpy (initpath + strlen ("/lib/i686/"), d->d_name);
		   unlink (initpath);
		   break;
		 }
	     }
	 base += d->d_reclen;
       }
      close (fd);
    }
#endif
  pid = vfork ();
  if (pid == 0) {
    execl ("/sbin/ldconfig", "/sbin/ldconfig", NULL);
    _exit (1);
  } else if (pid < 0) {
    _exit (1);
  }
  if (waitpid (0, &status, 0) != pid || !WIFEXITED (status)) {
    _exit (1);
  }
  if (WEXITSTATUS (status))
    _exit (WEXITSTATUS (status));

  if (! utimes (GCONV_MODULES_CACHE, NULL)) {
    pid = vfork ();
    if (pid == 0) {
      execl ("/usr/sbin/iconvconfig", "/usr/sbin/iconvconfig", NULL);
      _exit (1);
    } else if (pid < 0) {
      _exit (1);
    }
    if (waitpid (0, &status, 0) != pid || !WIFEXITED (status)) {
      _exit (1);
    }
    if (WEXITSTATUS (status))
      _exit (WEXITSTATUS (status));
  }

  /* Check if telinit is available and the init fifo as well.  */
  if (access ("/sbin/telinit", X_OK) || access ("/dev/initctl", F_OK))
    _exit (0);
  /* Check if we are not inside of some chroot, because we'd just
     timeout and leave /etc/initrunlvl.  */
  if (readlink ("/proc/1/exe", initpath, 256) <= 0 ||
      readlink ("/proc/1/root", initpath, 256) <= 0)
    _exit (0);
  pid = vfork ();
  if (pid == 0) {
    execl ("/sbin/telinit", "/sbin/telinit", "u", NULL);
    _exit (1);
  } else if (pid < 0) {
    _exit (1);
  }
  if (waitpid (0, &status, 0) != pid || !WIFEXITED (status)) {
    _exit (1);
  }
  _exit(0);
}

#ifndef __powerpc__
int __libc_start_main (int (*main) (void), int argc, char **argv,
		       void (*init) (void), void (*fini) (void),
		       void (*rtld_fini) (void), void * stack_end)
#else
struct startup_info
{
  void *sda_base;
  int (*main) (int, char **, char **, void *);
  int (*init) (int, char **, char **, void *);
  void (*fini) (void);
};

int __libc_start_main (int argc, char **ubp_av,
		       char **ubp_ev,
		       void *auxvec, void (*rtld_fini) (void),
		       struct startup_info *stinfo,
		       char **stack_on_entry)
#endif
{
  main();
  return 0;
}
